/*
Copyright (C) THL A29 Limited, a Tencent company. All rights reserved.

SPDX-License-Identifier: Apache-2.0
*/

// Package event provides a Redis client to publish and subscribe to events.
package event

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"strconv"
	"strings"

	"chainmaker.org/chainmaker/pb-go/v2/common"
	"github.com/go-redis/redis/v8"
)

// RedisEventKey is the key used to store the event data in Redis.
const RedisEventKey = "event"
const BlockHeightStr = "block_height"
const RedisAlreadyExists = "Consumer Group name already exists"

// RedisClient is a simple struct to hold the Redis client.
type RedisClient struct {
	redisClient *redis.Client
}

// NewRedisClient creates a new RedisClient with a connected Redis client.
//
//	@Description:
//	@param redisAddr
//	@param username
//	@param password
//	@return *RedisClient
//	@return error
func NewRedisClient(redisAddr, username, password string) (*RedisClient, error) {
	ctx := context.Background()
	rdb := redis.NewClient(&redis.Options{
		Addr:     redisAddr, // Redis地址
		Username: username,  // 用户名
		Password: password,  // 密码，没有则留空
		DB:       0,         // 使用默认DB
	})

	_, err := rdb.Ping(ctx).Result()
	if err != nil {
		return nil, err
	}

	return &RedisClient{
		redisClient: rdb,
	}, nil
}

// PublishEventToStream publishes an event to a Redis channel.
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param event
//	@param subChainID
//	@return error
func (p *RedisClient) PublishEventToStream(ctx context.Context, event *common.ContractEventInfo,
	subChainID string, chainInfoId int) error {
	message, err := json.Marshal(event)
	if err != nil {
		return fmt.Errorf("failed to marshal event, %v", err)
	}
	_, err = p.redisClient.XAdd(ctx, &redis.XAddArgs{
		Stream: strings.Join([]string{fmt.Sprintf("%d", chainInfoId), event.ChainId, subChainID, event.ContractName}, "#"),
		Values: map[string]interface{}{RedisEventKey: message},
	}).Result()
	if err != nil {
		return fmt.Errorf("failed to publish event stream to redis, %v", err)
	}
	return nil
}

// CreateConsumerGroup  creates a new consumer group in Redis.
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param streamId
//	@param groupName
//	@return error
func (p *RedisClient) CreateConsumerGroup(ctx context.Context, streamId, groupName string) error {
	_, err := p.redisClient.XGroupCreateMkStream(ctx, streamId, groupName, "0").Result()
	if err != nil && !strings.Contains(err.Error(), RedisAlreadyExists) {
		return err
	}
	return nil
}

// SubscribeToStream subscribes to a Redis stream and processes events using the provided handler function.
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param chainID
//	@param subChainID
//	@param contractName
//	@param eventName
//	@param groupName
//	@param consumerName
//	@param handler
//	@return error
func (p *RedisClient) SubscribeToStream(ctx context.Context, chainID, subChainID, contractName,
	groupName, consumerName string, chainInfoId int, handler func(*common.ContractEventInfo)) error {
	streamId := strings.Join([]string{fmt.Sprintf("%d", chainInfoId), chainID, subChainID, contractName}, "#")
	if err := p.CreateConsumerGroup(ctx, streamId, groupName); err != nil {
		return fmt.Errorf("failed to create consumer group, %v", err)
	}
	for {
		xReadGroupArgs := &redis.XReadGroupArgs{
			Group:    groupName,
			Consumer: consumerName,
			Streams:  []string{streamId, ">"},
			Count:    1,
			Block:    0,
			NoAck:    false,
		}
		messages, err := p.redisClient.XReadGroup(ctx, xReadGroupArgs).Result()
		if err != nil {
			if errors.Is(err, redis.Nil) {
				continue
			}
			return fmt.Errorf("failed to read from stream, %v", err)
		}

		for _, message := range messages {
			for _, xMessage := range message.Messages {
				eventData, ok := xMessage.Values[RedisEventKey]
				if !ok {
					continue
				}

				eventDataBytes, ok := eventData.(string)
				if !ok {
					continue
				}

				var event common.ContractEventInfo
				if err = json.Unmarshal([]byte(eventDataBytes), &event); err != nil {
					return fmt.Errorf("failed to unmarshal event data, %v", err)
				}

				handler(&event)

				_, err = p.redisClient.XAck(ctx, streamId, groupName, xMessage.ID).Result()
				if err != nil {
					return fmt.Errorf("failed to ack message, %v", err)
				}
			}
		}
	}
}

// SubscribeToStreamWithHandlerError subscribes to a Redis stream and processes events using
// the provided handler function, and catch handler error.
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param chainID
//	@param subChainID
//	@param contractName
//	@param eventName
//	@param groupName
//	@param consumerName
//	@param handler
//	@return error
func (p *RedisClient) SubscribeToStreamWithHandlerError(ctx context.Context, chainID, subChainID, contractName,
	groupName, consumerName string, chainInfoId int, handler func(*common.ContractEventInfo) error) error {
	streamId := strings.Join([]string{fmt.Sprintf("%d", chainInfoId), chainID, subChainID, contractName}, "#")
	if err := p.CreateConsumerGroup(ctx, streamId, groupName); err != nil {
		return fmt.Errorf("failed to create consumer group, %v", err)
	}
	for {
		xReadGroupArgs := &redis.XReadGroupArgs{
			Group:    groupName,
			Consumer: consumerName,
			Streams:  []string{streamId, ">"},
			Count:    1,
			Block:    0,
			NoAck:    false,
		}
		messages, err := p.redisClient.XReadGroup(ctx, xReadGroupArgs).Result()
		if err != nil {
			if errors.Is(err, redis.Nil) {
				continue
			}
			return fmt.Errorf("failed to read from stream, %v", err)
		}

		for _, message := range messages {
			for _, xMessage := range message.Messages {
				eventData, ok := xMessage.Values[RedisEventKey]
				if !ok {
					continue
				}

				eventDataBytes, ok := eventData.(string)
				if !ok {
					continue
				}

				var event common.ContractEventInfo
				if err = json.Unmarshal([]byte(eventDataBytes), &event); err != nil {
					return fmt.Errorf("failed to unmarshal event data, %v", err)
				}

				if err = handler(&event); err != nil {
					return fmt.Errorf("failed to handler event, %v", err)
				}

				_, err = p.redisClient.XAck(ctx, streamId, groupName, xMessage.ID).Result()
				if err != nil {
					return fmt.Errorf("failed to ack message, %v", err)
				}
			}
		}
	}
}

// SetLatestBlockHeight sets the latest block height
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param chainID
//	@param height
//	@return error
func (p *RedisClient) SetLatestBlockHeight(ctx context.Context, chainID string, height int64) error {
	err := p.redisClient.Set(ctx, strings.Join([]string{BlockHeightStr, chainID}, "#"),
		strconv.FormatInt(height, 10), 0).Err()
	if err != nil {
		return fmt.Errorf("failed to set redis kv, %v", err)
	}
	return nil
}

// GetLatestBlockHeight gets the latest block height
//
//	@Description:
//	@receiver p
//	@param ctx
//	@param chainID
//	@return int64
//	@return error
func (p *RedisClient) GetLatestBlockHeight(ctx context.Context, chainID string) (int64, error) {
	val, err := p.redisClient.Get(ctx, strings.Join([]string{BlockHeightStr, chainID}, "#")).Result()
	if err != nil {
		if errors.Is(err, redis.Nil) {
			return 0, nil
		}
		return -1, fmt.Errorf("failed to get redis kv, %v", err)
	}
	height, err := strconv.ParseInt(val, 10, 64)
	if err != nil {
		return -1, fmt.Errorf("failed to atoi string[%s], %v", val, err)
	}
	return height, nil
}
